Add LoadPath "CoqEAL".
Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype.
Require Import finfun tuple ssralg matrix mxalgebra bigop.
Require Import ZArith Zinfra definitions_auxlemmas.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


(*In this file, we have a reduction rho from the chain complex D_rho to d_rho*)


Section decomposition.

Open Local Scope ring_scope.

Variable K : fieldType.
Variable rho : FGReduction K.

Notation "'D_rho'" := (diff (C rho)).
Notation "'d_rho'" := (diff (D rho)).
Notation "'F_rho'" := (M (F rho)).
Notation "'G_rho'" := (M (G rho)).
Notation "'H_rho'" := (Ho (H rho)).
Notation "'ci'" := (m (C rho)).

Section definitions.


(*We want to define an isomorphism between the chain complex D_rho and another one
decomposed *)

Section isomorphism.


(*We define the morphisms Fi_isom and Gi_isom which will let us decompose every chain group in three parts:
   - ker_min F_rho i intersection ker_min H_rho (i+1)
   - ker_min F_rho (i+1) intersection ker_min D_rho i
   - im G_rho (i+1)

Fi_isom goes from D_rho to the decomposed one.
Gi_isom goes from the decomposed to D_rho *)

Definition castFiDi (i : Z) := 
(castmx ((cast_rank1 (row_mx (F_rho (i+1)) (D_rho i))),
                     (Logic.eq_refl 
              (ci (i+1) - \rank (row_mx (F_rho (i+1)) (D_rho i)))%nat))
    (col_mx 
       (@const_mx K (\rank (row_mx (F_rho (i+1)) (D_rho i))) 
          ((ci (i+1))-\rank (row_mx (F_rho (i+1)) (D_rho i)))%nat 0) 1%:M)).

Definition castFiHi (i : Z) := 
(castmx ((cast_rank1 (row_mx (F_rho i) (H_rho i))),
                     (Logic.eq_refl ((ci i)-\rank (row_mx (F_rho i) (H_rho i)))%nat))
    (col_mx 
       (@const_mx K (\rank (row_mx (F_rho i) (H_rho i))) ((ci i)-\rank (row_mx (F_rho i) (H_rho i)))%nat 0) 1%:M)).


Definition Fi_isom (i : Z):= 
   (col_mx (ker_min (row_mx (F_rho (i+1)) (H_rho (i+1)))) 
           (col_mx (ker_min (row_mx (F_rho (i+1)) (D_rho i)))
             (G_rho (i+1)))).


Definition Gi_isom (i : Z) := 
  (row_mx ((D_rho i) *m (H_rho i) *m
              (col_ebase (row_mx (F_rho (i+1)) (H_rho (i+1))))
                                *m (castFiHi (i+1))) 
    (row_mx ((H_rho (i+1)) *m (D_rho (i+1)) *m 
              (col_ebase (row_mx (F_rho (i+1)) (D_rho i)))
               *m (castFiDi i))
            (F_rho (i+1)))).

Section isomorphism.

(*Exists an isomorphism between Gi_isom and Fi_isom: 
  GiFi = Id and FiGi = Id *)

Lemma GiFi (i : Z) : (Gi_isom i) *m (Fi_isom i) = 1%:M.
Proof.
rewrite /Fi_isom /Gi_isom !mul_row_col -mulmxA /castFiDi -!mulmxA.
have -> : (castmx  (cast_rank1 (row_mx  (F_rho (i + 1))  (D_rho i)),
         erefl  (ci (i + 1) -  \rank   (row_mx  (F_rho (i + 1))  (D_rho i)))%nat)
         (col_mx (const_mx 0) 1%:M) *m ker_min (row_mx (F_rho (i + 1)) (D_rho i))) = 
         kermx (row_mx (F_rho (i + 1)) (D_rho i)).
  rewrite /ker_min mulmxA mulmx_cast ?cast_rank1 // => Hyp0.
  by rewrite mul_col_row !mul0mx mulmx0 mulmx1 /kermx castmx_blockmx_copid 
             mulmxA copid_mx_id ?(rank_leq_row (row_mx (F_rho (i + 1)) (D_rho i)) ). 
have H1 : H_rho (i + 1) *m (D_rho (i + 1)) *m (row_mx (F_rho (i + 1)) (D_rho i)) = 0.
  by rewrite mul_mx_row -mulmxA -mulmxA ((nilpotency (C rho)) i) mulmx0
          ((M_well_defined (F rho)) (i+1)) mulmxA ((ax3 rho) (i+1))
          mul0mx row_mx0.
have -> : H_rho (i + 1) *m (D_rho (i + 1) *m (col_ebase
                                          (row_mx (F_rho (i + 1)) (D_rho i)) *m 
                                        kermx
                                          (row_mx (F_rho (i + 1)) (D_rho i)))) = (H_rho (i + 1) *m D_rho (i + 1) *m 
                               col_ebase (row_mx (F_rho (i + 1)) (D_rho i)) *m 
                               kermx (row_mx (F_rho (i + 1)) (D_rho i)))
  by rewrite !mulmxA.
rewrite  (mulmxKV_ker H1) /castFiHi. 
have -> : (castmx (cast_rank1 (row_mx (F_rho (i + 1)) (H_rho (i + 1))),
           erefl (ci (i + 1) -  \rank (row_mx (F_rho (i + 1))
           (H_rho (i + 1))))%nat)  (col_mx (const_mx 0) 1%:M) *m 
           ker_min (row_mx (F_rho (i + 1)) (H_rho (i + 1)))) = 
           kermx (row_mx (F_rho (i + 1)) (H_rho (i + 1))).
  rewrite /ker_min mulmxA mulmx_cast ?cast_rank1 // => Hyp0.
  by rewrite mul_col_row !mul0mx mulmx0 mulmx1 /kermx castmx_blockmx_copid 
             mulmxA copid_mx_id ?(rank_leq_row (row_mx (F_rho (i + 1)) (H_rho (i + 1)))).
have HH2 : D_rho i *m H_rho i *m (row_mx (F_rho (i + 1)) (H_rho (i + 1)))  = 0.
  by rewrite mul_mx_row -mulmxA -mulmxA ((ax5 rho) i) mulmx0 ((ax3 rho) i)
          mulmx0 row_mx0.
have -> : D_rho i *m (H_rho i *m (col_ebase (row_mx (F_rho (i + 1)) (H_rho (i + 1))) *m 
                           kermx (row_mx (F_rho (i + 1)) (H_rho (i + 1))))) =
D_rho i *m H_rho i *m col_ebase (row_mx (F_rho (i + 1)) (H_rho (i + 1))) *m 
                           kermx (row_mx (F_rho (i + 1)) (H_rho (i + 1)))
  by rewrite !mulmxA. 
by rewrite (mulmxKV_ker HH2) GRing.addrA GRing.addrC
       [D_rho i *m H_rho i + _]GRing.addrC GRing.addrA (ax2 rho).
Qed.




Lemma FiGi (i : Z) : (Fi_isom i) *m (Gi_isom i) = 1%:M.
Proof.
rewrite /Gi_isom /Fi_isom !mul_col_row  (ax1 rho (i+1)) !mul_mx_row mulmxA mulmxA
        mulmxA. 
have ->: (G_rho (i + 1) *m (H_rho (i + 1) *m D_rho (i + 1) *m 
                           col_ebase (row_mx (F_rho (i + 1)) (D_rho i)) *m 
                           castFiDi i)) = 
         (G_rho (i + 1) *m H_rho (i + 1) *m D_rho (i + 1) *m 
                           col_ebase (row_mx (F_rho (i + 1)) (D_rho i)) *m 
                           castFiDi i).
  by rewrite !mulmxA.
rewrite (ax4 rho (i+1)) !mul0mx mulmxA mulmxA mul_col_mx mulmxA.
have -> : (G_rho (i + 1) *m (D_rho i *m H_rho i *m col_ebase
                                                  (row_mx 
                                                  (F_rho (i + 1))
                                                  (H_rho (i + 1))) *m 
                           castFiHi (i + 1))) = 
   (G_rho (i + 1) *m D_rho i *m H_rho i *m col_ebase
                                                  (row_mx 
                                                  (F_rho (i + 1))
                                                  (H_rho (i + 1))) *m 
                           castFiHi (i + 1)).
  by rewrite !mulmxA.
rewrite  -((M_well_defined (G rho)) (i)) -[_ *m (G_rho i) *m _]mulmxA (ax4 rho i) mulmx0 !mul0mx.
move : (eqmx_trans (ker_intersection (F_rho (i+1))(D_rho i)) (ker_min_kermx (row_mx (F_rho (i+1))(D_rho i)))).
move/eqmxP; move/andP => [_ Hb].
move : Hb; move/submxP => [D' HD']; rewrite HD'.
move : (capmxMr (kermx (F_rho (i+1))) (kermx (D_rho i))(F_rho (i+1))).
move/submxP => [D HD].
rewrite -[D' *m _ *m (F_rho (i+1))]mulmxA HD mulmx_ker cap0mx !mulmx0.
move : (eqmx_trans (ker_intersection (F_rho (i+1)) (H_rho (i+1))) (ker_min_kermx (row_mx  (F_rho (i+1)) (H_rho (i+1))))).
move/eqmxP; move/andP => [_ Hb].
move : Hb; move/submxP => [D'' HD'']; rewrite HD''.
move : (capmxMr (kermx (F_rho (i+1))) (kermx (H_rho (i+1))) (F_rho (i+1))).
move/submxP => [Da HDa].
rewrite -[D'' *m _ *m  (F_rho (i+1))]mulmxA HDa mulmx_ker cap0mx !mulmx0. 
have ->: (D' *m (kermx (F_rho (i + 1)) :&: kermx (D_rho i))%MS *m 
         (D_rho i *m H_rho i *m col_ebase
                                  (row_mx (F_rho (i + 1)) (H_rho (i + 1))) *m 
          castFiHi (i + 1))) = 0.
  rewrite mulmxA mulmxA mulmxA -[D' *m _ *m (D_rho i)]mulmxA.
  move : (capmxMr (kermx (F_rho (i + 1))) (kermx (D_rho i)) (D_rho i)).
  move/submxP => [Db ->].
  by rewrite mulmx_ker capmx0 !mulmx0 !mul0mx.
have -> : (D'' *m (kermx (F_rho (i + 1)) :&: kermx (H_rho (i+1)))%MS *m (H_rho (i+1)) *m (D_rho (i+1)) *m 
         col_ebase (row_mx (F_rho (i + 1)) (D_rho i)) *m (castFiDi i)) =
  (D'' *m (kermx (F_rho (i+1)) :&: kermx (H_rho (i+1)))%MS *m ((H_rho (i+1)) *m (D_rho (i+1)) *m col_ebase (row_mx (F_rho (i+1)) (D_rho i)) *m (castFiDi i))) by rewrite !mulmxA.
have -> : (D'' *m (kermx (F_rho (i+1)) :&: kermx (H_rho (i+1)))%MS *m ((H_rho (i+1)) *m (D_rho (i+1)) *m col_ebase (row_mx (F_rho (i+1)) (D_rho i)) *m (castFiDi i))) = 0.
  rewrite mulmxA mulmxA mulmxA -[D'' *m _ *m (H_rho (i+1))]mulmxA.
  move : (capmxMr (kermx (F_rho (i+1))) (kermx (H_rho (i+1))) (H_rho (i+1))).
  move/submxP => [Db ->].
  by rewrite mulmx_ker capmx0 !mulmx0 !mul0mx.
have -> : (D' *m (kermx (F_rho (i+1)) :&: kermx (D_rho i))%MS *m ((H_rho (i+1)) *m (D_rho (i+1)) *m 
                                              col_ebase 
                                                (row_mx (F_rho (i+1)) (D_rho i)) *m (castFiDi i))) = 1%:M.
  have -> : ((H_rho (i+1)) *m (D_rho (i+1))) = 1%:M - ((D_rho i) *m (H_rho i)) - ((F_rho (i+1)) *m (G_rho (i+1))).
    rewrite -(ax2 rho i). 
    have -> : (F_rho (i+1)) *m (G_rho (i+1)) + (H_rho (i+1)) *m (D_rho (i+1)) + (D_rho i) *m (H_rho i) - (D_rho i) *m (H_rho i) - (F_rho (i+1)) *m (G_rho (i+1)) =
      (F_rho (i+1)) *m (G_rho (i+1)) + (H_rho (i+1)) *m (D_rho (i+1)) + ((D_rho i) *m (H_rho i) - (D_rho i) *m (H_rho i)) - (F_rho (i+1)) *m (G_rho (i+1))
      by rewrite GRing.addrA. 
    have H4 : (D_rho i) *m (H_rho i) == (D_rho i) *m (H_rho i) by done.
    rewrite -GRing.subr_eq0 in H4.
    rewrite (eqP H4) GRing.addr0.
    rewrite [(_ *m_) + (_ *m _)]GRing.addrC -GRing.addrA.
    clear H4.
    have H4 : (F_rho (i+1)) *m (G_rho (i+1)) == (F_rho (i+1)) *m (G_rho (i+1)) by done.
    rewrite -GRing.subr_eq0 in H4.
    by rewrite (eqP H4) GRing.addr0.
  rewrite !mulmx_addl !mulmx_addr mul1mx !mulNmx !mulmxN.
  have -> : D' *m (kermx (F_rho (i+1)) :&: kermx (D_rho i))%MS *m ((D_rho i) *m (H_rho i) *m 
                                        col_ebase (row_mx (F_rho (i+1)) (D_rho i)) *m (castFiDi i)) = 0.
    rewrite mulmxA mulmxA mulmxA -[D' *m _ *m (D_rho i)]mulmxA.
    move : (capmxMr (kermx (F_rho (i+1))) (kermx (D_rho i)) (D_rho i)).
    move/submxP => [Db ->].
    by rewrite mulmx_ker capmx0 !mulmx0 !mul0mx.
  rewrite GRing.subr0.
  have -> : (D' *m (kermx (F_rho (i+1)) :&: kermx (D_rho i))%MS *m ((F_rho (i+1)) *m (G_rho (i+1)) *m col_ebase (row_mx (F_rho (i+1)) (D_rho i)) *m (castFiDi i))) = 0.
    rewrite mulmxA mulmxA mulmxA -[D' *m _ *m (F_rho (i+1))]mulmxA.
    move : (capmxMr (kermx (F_rho (i+1))) (kermx (D_rho i)) (F_rho (i+1))).
    move/submxP => [Db ->].
    by rewrite mulmx_ker cap0mx !mulmx0 !mul0mx.
  rewrite GRing.subr0.
  rewrite -HD' /ker_min /kermx -!mulmxA mulmxA. 
  have -> : 
   (invmx (col_ebase (row_mx (F_rho (i+1)) (D_rho i))) *m (col_ebase (row_mx (F_rho (i+1)) (D_rho i)) 
         *m (castFiDi i))) = (castFiDi i). 
  by rewrite mulmxA (mulVmx (col_ebase_unit (row_mx (F_rho (i+1)) (D_rho i)))) mul1mx. 
  rewrite -(@castmx_blockmx_copid _ ((ci (i + 1)) - \rank (row_mx (F_rho (i+1)) (D_rho i)))) ?(cast_rank1 (row_mx (F_rho (i+1)) (D_rho i))) // => Hyp0; rewrite mulmx_cast ?(cast_rank1 (row_mx (F_rho (i+1)) (D_rho i))) // => Hyp1; rewrite mulmx_cast ?(cast_rank1 (row_mx (F_rho (i+1)) (D_rho i))) // => Hyp2.
  by rewrite mul_row_block mul_row_col mul0mx GRing.add0r mulmx0 GRing.add0r
          !mulmx1 mulmx0 GRing.add0r castmx_id.
have -> : (D'' *m (kermx (F_rho (i+1)) :&: kermx (H_rho (i+1)))%MS *m (D_rho i) *m (H_rho i) *m 
      col_ebase (row_mx (F_rho (i+1)) (H_rho (i+1))) *m (castFiHi (i+1)))  =
(D'' *m (kermx (F_rho (i+1)) :&: kermx (H_rho (i+1)))%MS *m ((D_rho i) *m (H_rho i) *m 
                                              col_ebase 
                                                (row_mx (F_rho (i+1)) (H_rho (i+1))) *m  (castFiHi (i+1))))
by rewrite !mulmxA.
have -> : (D'' *m (kermx (F_rho (i+1)) :&: kermx (H_rho (i+1)))%MS *m ((D_rho i) *m (H_rho i) *m 
                                              col_ebase 
                                                (row_mx (F_rho (i+1)) (H_rho (i+1))) *m (castFiHi(i+1)))) = 1%:M.
  have -> : ((D_rho i) *m (H_rho i)) = 1%:M - ((H_rho (i+1)) *m (D_rho (i+1))) - ((F_rho (i+1)) *m (G_rho (i+1))).
    rewrite -(ax2 rho i) -[(F_rho (i+1)) *m (G_rho (i+1)) + _ + _]GRing.addrA [(H_rho (i+1)) *m _ + _]GRing.addrC
      GRing.addrA. 
    have -> : (F_rho (i+1)) *m (G_rho (i+1)) + (D_rho i) *m (H_rho i) + (H_rho (i+1)) *m (D_rho (i+1)) - (H_rho (i+1)) *m (D_rho (i+1)) - (F_rho (i+1)) *m (G_rho (i+1)) =
    (F_rho (i+1)) *m (G_rho (i+1)) + (D_rho i) *m (H_rho i) + ((H_rho (i+1)) *m (D_rho (i+1)) - (H_rho (i+1)) *m (D_rho (i+1))) - (F_rho (i+1)) *m (G_rho (i+1))
    by rewrite GRing.addrA. 
    have H4 : (H_rho (i+1)) *m (D_rho (i+1)) == (H_rho (i+1)) *m (D_rho (i+1)) by done.
    rewrite -GRing.subr_eq0 in H4.
    rewrite (eqP H4) GRing.addr0.
    rewrite [(_ *m_) + (_ *m _)]GRing.addrC -GRing.addrA.
    clear H4.
    have H4 : (F_rho (i+1)) *m (G_rho (i+1)) == (F_rho (i+1)) *m (G_rho (i+1)) by done.
    rewrite -GRing.subr_eq0 in H4.
    by rewrite (eqP H4) GRing.addr0.  
    rewrite !mulmx_addl !mulmx_addr mul1mx !mulNmx !mulmxN.
  have -> : D'' *m (kermx (F_rho (i+1)) :&: kermx (H_rho (i+1)))%MS *m ((H_rho (i+1)) *m (D_rho (i+1)) *m 
                                        col_ebase (row_mx (F_rho (i+1)) (H_rho (i+1))) *m  (castFiHi (i+1))) = 0.
    rewrite mulmxA mulmxA mulmxA -[D'' *m _ *m (H_rho (i+1))]mulmxA.
    move : (capmxMr (kermx (F_rho (i+1))) (kermx (H_rho (i+1))) (H_rho (i+1))).
    move/submxP => [Db ->].
    by rewrite mulmx_ker capmx0 !mulmx0 !mul0mx.
  rewrite GRing.subr0.
  have -> : (D'' *m (kermx (F_rho (i+1)) :&: kermx (H_rho (i+1)))%MS *m ((F_rho (i+1)) *m (G_rho (i+1)) *m col_ebase (row_mx (F_rho (i+1)) (H_rho (i+1))) *m (castFiHi (i+1)))) = 0.
    rewrite mulmxA mulmxA mulmxA -[D'' *m _ *m (F_rho (i+1))]mulmxA.
    move : (capmxMr (kermx (F_rho (i+1))) (kermx (H_rho (i+1))) (F_rho (i+1))).
    move/submxP => [Db ->].
    by rewrite mulmx_ker cap0mx !mulmx0 !mul0mx.
  rewrite GRing.subr0.
  rewrite -HD'' /ker_min /kermx -!mulmxA mulmxA. 
  have -> : 
   (invmx (col_ebase (row_mx (F_rho (i+1)) (H_rho (i+1)))) *m (col_ebase (row_mx (F_rho (i+1)) (H_rho (i+1))) 
         *m  (castFiHi (i+1)))) = castFiHi (i+1) 
  by rewrite mulmxA (mulVmx (col_ebase_unit (row_mx (F_rho (i+1)) (H_rho (i+1))))) mul1mx. 
  rewrite -(@castmx_blockmx_copid _ ((ci (i+1)) - \rank (row_mx (F_rho (i+1)) (H_rho (i+1))))) ?(cast_rank1 (row_mx (F_rho (i+1)) (H_rho (i+1)))) // => Hyp0; rewrite mulmx_cast ?(cast_rank1 (row_mx (F_rho (i+1)) (H_rho (i+1)))) // => Hyp1; rewrite mulmx_cast ?(cast_rank1 (row_mx (F_rho (i+1)) (H_rho (i+1)))) // => Hyp2.
  by rewrite mul_row_block mul_row_col mul0mx GRing.add0r mulmx0 GRing.add0r
          !mulmx1 mulmx0 GRing.add0r castmx_id.
by rewrite -scalar_mx_block row_mx0 col_mx0 -scalar_mx_block.
Qed.

End isomorphism.


(*We are going to build a reduction between the decomposed chain complex and d_rho.

Firstly, we define the differentials and the homotopy operator of the decomposed chain complex and we prove that these newe differentials define the decomposed chain complex, newC_change_base *)

Section new_D_new_H.

Definition  D_rho_change_base (i:Z):=
 (Fi_isom (i+1) *m (D_rho (i+1)) *m (Gi_isom i)).

Definition H_rho_change_base (i:Z):= 
 (Fi_isom i) *m (H_rho (i+1)) *m (Gi_isom (i+1)).


Lemma di_plus_1di : forall i:Z, (D_rho_change_base (i + 1)) *m (D_rho_change_base i) = 0.
Proof. 
 rewrite /D_rho_change_base. 
move=> i0. rewrite -mulmxA. 
by rewrite ![(Gi_isom (i0 + 1) *m _)]mulmxA GiFi mul1mx -mulmxA
          [(D_rho (i0+1+1)) *m _]mulmxA (nilpotency (C rho) (i0+1)) mul0mx mulmx0.
Qed.

Definition newC_change_base:= Build_FGChain_Complex di_plus_1di.

(*Moreover, we define the morphisms between the decomposed chain complex (newC_change_base) and d_rho. These are defined by the composition Fi_isom and F_rho (Gi_isom and G_rho).

With these defnitions, we can prove the properties to build a reduction.
*)


Definition F_rho_change_base (i:Z) := 
 Fi_isom (i) *m (F_rho (i+1)).


Definition G_rho_change_base (i:Z) := 
 G_rho (i+1)  *m Gi_isom (i).


Section reduction_properties.


(*The five lemmas necessary to define a reduction between the descomposed chain
complex and the reduced one*)

Lemma fg_change_base  :forall i:Z, G_rho_change_base i *m F_rho_change_base i = 1%:M.
Proof.
move => i. 
by rewrite /(G_rho_change_base i) /(F_rho_change_base i) -mulmxA [(Gi_isom i) *m _]mulmxA (GiFi i) mul1mx (ax1 rho (i+1)).
Qed.

Lemma fh_change_base :forall i:Z,  H_rho_change_base i *m F_rho_change_base (i+1) = 0.
Proof.
move => i.
by rewrite /H_rho_change_base /F_rho_change_base -mulmxA [(Gi_isom (i+1)) *m _]mulmxA (GiFi (i+1)) mul1mx 
          -mulmxA (ax3 rho (i+1)) mulmx0.
Qed.

Lemma hg_change_base : forall i:Z,  G_rho_change_base i *m H_rho_change_base i = 0.
Proof.
move => i.
by rewrite /G_rho_change_base /H_rho_change_base -mulmxA ![(Gi_isom i) *m _]mulmxA (GiFi i) mul1mx mulmxA
           (ax4 rho (i+1)) mul0mx.
Qed.

Lemma hh_change_base : forall i:Z,  H_rho_change_base i *m H_rho_change_base (i+1) = 0.
Proof.
move => i.
by rewrite /H_rho_change_base -mulmxA ![(Gi_isom (i+1)) *m _]mulmxA (GiFi (i+1)) mul1mx -mulmxA
           [(H_rho (i+1)) *m _]mulmxA (ax5 rho (i+1)) mul0mx mulmx0.
Qed.


Lemma gf_Dh_hD_change_base : forall i:Z, 
  F_rho_change_base (i+1) *m G_rho_change_base(i+1) + H_rho_change_base (i+1) *m D_rho_change_base(i+1) + D_rho_change_base i *m H_rho_change_base i = 1%:M.
Proof.
move => i. 
rewrite /G_rho_change_base /F_rho_change_base /D_rho_change_base /H_rho_change_base -!mulmxA
        -mulmx_addr !mulmxA -mulmx_addl -!mulmxA -mulmx_addr
        !mulmxA -mulmx_addl.
have -> : H_rho (i+1+1) *m Gi_isom (i+1+1) *m Fi_isom (i+1+1) *m D_rho(i+1+1) = H_rho (i+1+1) *m (Gi_isom (i+1+1) *m Fi_isom (i+1+1)) *m D_rho(i+1+1)
  by rewrite !mulmxA.
rewrite (GiFi (i+1+1)) mulmx1.
have -> :   D_rho (i + 1) *m Gi_isom i *m 
                        Fi_isom i *m H_rho (i + 1)= D_rho (i + 1) *m (Gi_isom i *m 
                        Fi_isom i) *m H_rho (i + 1).
  by rewrite !mulmxA.
by rewrite (GiFi i) mulmx1 (ax2 rho (i+1)) mul1mx (FiGi (i+1)).
Qed.



(*Proofs: F_rho_change_base and G_rho_change_base is a chain complex morphism*)

Lemma d1f1_f0D1 :forall i:Z, F_rho_change_base (i+1) *m (d_rho (i+1)) = (D_rho_change_base i) *m F_rho_change_base i.
Proof. 
move => i.
by rewrite /D_rho_change_base /F_rho_change_base -!mulmxA [Gi_isom i *m _]mulmxA (GiFi i) mul1mx ((M_well_defined (F rho))(i+1)).
Qed.

Lemma D1g1_g0d1 :forall i:Z, G_rho_change_base (i+1) *m D_rho_change_base (i) = (d_rho (i+1)) *m G_rho_change_base i.
Proof. 
move => i.
by rewrite /D_rho_change_base /G_rho_change_base -!mulmxA [Gi_isom (i+1) *m _]mulmxA (GiFi (i+1)) mul1mx !mulmxA ((M_well_defined (G rho))(i+1)).
Qed.

End reduction_properties.



(*Up to now, we have a reduction between the decomposed chain complex and d_rho *)

(*Concretely, we know how are the matrices: F_rho_change_base,G_rho_change_base, 
  H_rho_change_base and D_rho_change_base*)


 Lemma F_rho_change_baseP : forall i: Z,
  F_rho_change_base i = 
  (col_mx 0 (col_mx 0  1%:M)).
Proof.
move=> i.
rewrite /F_rho_change_base /Fi_isom !mul_col_mx.  
move : 
  (eqmx_trans (ker_intersection (F_rho (i+1)) (H_rho (i+1))) (ker_min_kermx (row_mx (F_rho (i+1)) (H_rho (i+1))))).
  move/eqmxP; move/andP => [_ Hb].
  move : Hb; move/submxP => [D'' HD'']; rewrite HD''.
  move : (capmxMr (kermx (F_rho (i+1))) (kermx (H_rho (i+1))) (F_rho (i+1))).
  move/submxP => [Da HDa].
 rewrite -[D'' *m _ *m (F_rho (i+1))]mulmxA.
 rewrite HDa.
 rewrite mulmx_ker cap0mx !mulmx0. 

move : 
  (eqmx_trans (ker_intersection (F_rho (i+1)) (D_rho i)) (ker_min_kermx (row_mx (F_rho (i+1)) (D_rho i)))).
  move/eqmxP; move/andP => [_ Hb].
  move : Hb; move/submxP => [D1'' HD1'']; rewrite HD1''.
  move : (capmxMr (kermx (F_rho (i+1))) (kermx (D_rho i)) (F_rho (i+1))).
  move/submxP => [Da' HDa'].
  rewrite -[D1'' *m _ *m (F_rho (i+1))]mulmxA.
 rewrite HDa'.
by rewrite mulmx_ker cap0mx !mulmx0 (ax1 rho).
Qed.

(*
Definition G_aux (i:Z) := G_rho (i+1) *m F_rho (i + 1).*)

 Lemma G_rho_change_baseP : forall i: Z,
  G_rho_change_base i = 
  (row_mx 0 (row_mx 0  1%:M)).
Proof.
move=> i.
by rewrite /G_rho_change_base !mul_mx_row mulmxA  mulmxA mulmxA -((M_well_defined (G rho))(i)) -[_ *m (G_rho i)  *m _]mulmxA (ax4 rho i) mulmx0 !mul0mx mulmxA mulmxA mulmxA (ax4 rho (i+1)) !mul0mx (ax1 rho). 
Qed.




Definition H_aux (i:Z) := (ker_min (row_mx (F_rho (i + 1)) (D_rho i)) *m 
            H_rho (i + 1) *m D_rho (i + 1) *m H_rho (i + 1) *m 
            col_ebase (row_mx (F_rho (i + 1 + 1)) (H_rho (i + 1 + 1))) *m 
            castFiHi (i + 1 + 1)).


Lemma H_rho_change_baseP : forall i: Z,
  H_rho_change_base i = 
  block_mx 0 (row_mx 0 0) (col_mx (H_aux i) 0) (block_mx 0 0 0 0) .
Proof.
move=> i.
rewrite /H_rho_change_base !mul_col_mx mul_mx_row mul_mx_row /(Gi_isom) (ax4 rho (i+1)) !mul0mx -[_ *m F_rho (i+1+1)]mulmxA (ax3 rho (i+1)) !mulmx0.
move : (eqmx_trans (ker_intersection (F_rho (i+1)) (H_rho (i+1))) (ker_min_kermx (row_mx (F_rho (i+1)) (H_rho (i+1))))).
move/eqmxP; move/andP => [_ Hb].
move : Hb; move/submxP => [D'' HD'']; rewrite HD''.
move : (capmxMr (kermx (F_rho (i+1))) (kermx (H_rho (i+1))) (H_rho (i+1))).
move/submxP => [Da HDa].
rewrite mulmxA -[D'' *m _ *m (H_rho (i+1))]mulmxA.
move: (capmxMr (kermx (F_rho (i+1))) (kermx (H_rho (i+1))) (H_rho (i+1))).
move/submxP => [Db ->].
by rewrite mulmx_ker capmx0 !mulmx0 !mul0mx !mul_mx_row !mulmxA -[_*m (H_rho (i+1)) *m (H_rho (i+1+1))]mulmxA (ax5 rho (i+1)) -[_ *m F_rho (i+1+1)]mulmxA (ax3 rho (i+1)) !mulmx0 !mul0mx -row_mx0 -row_mx0 -block_mxEv block_mxEh  -!block_mxEv.
Qed.


Definition D_aux i :=
 (ker_min (row_mx (F_rho (i+1+1)) (H_rho (i+1+1))) *m (D_rho (i+1)) *m
          ((H_rho (i+1)) *m (D_rho (i+1)) *m col_ebase (row_mx (F_rho (i+1)) (D_rho i)) *m (castFiDi i))).

Lemma D_rho_change_baseP : forall (i:Z), 
  D_rho_change_base i = 
  block_mx 0 (row_mx (D_aux i) 0) (col_mx 0 0) (block_mx 0 0 0 (d_rho (i+1))).
Proof.
move=> i.
rewrite /D_rho_change_base !mul_col_mx mul_mx_row mul_mx_row. 
rewrite /(Gi_isom (i+1)).
rewrite mul_mx_row.
rewrite !mul_mx_row -block_mxEv block_mxEh -!block_mxEv.
have -> : ((G_rho (i+1+1)) *m (D_rho (i+1)) *m (F_rho (i+1))) = (d_rho (i+1)).
  by rewrite -mulmxA ((M_well_defined (F rho)) (i+1)) mulmxA (ax1 rho (i+1 +1)) mul1mx.
have -> : ((G_rho (i+1+1)) *m (D_rho (i+1)) *m ((H_rho (i+1)) *m (D_rho (i+1)) *m col_ebase (row_mx (F_rho (i+1)) (D_rho i)) *m (castFiDi i))) = 0.
  by rewrite -((M_well_defined (G rho))(i+1)) mulmxA mulmxA mulmxA -[_ *m (G_rho (i+1)) *m _]mulmxA (ax4 rho (i+1)) mulmx0 !mul0mx.
have -> : (G_rho (i + 1 + 1) *m D_rho (i + 1) *m (D_rho i *m  H_rho i *m 
    col_ebase (row_mx  (F_rho (i + 1)) (H_rho (i + 1))) *m  castFiHi  (i + 1))) = 0.
  by rewrite mulmxA mulmxA mulmxA -[(G_rho (i+1+1)) *m _ *m _]mulmxA
    (nilpotency (C rho) i) mulmx0 !mul0mx.
have -> : (ker_min (row_mx (F_rho (i + 1 + 1)) (D_rho (i + 1))) *m 
    D_rho (i + 1) *m (D_rho i *m H_rho i *m col_ebase (row_mx   (F_rho (i + 1))
      (H_rho (i + 1))) *m  castFiHi (i + 1))) = 0.
  by rewrite mulmxA mulmxA mulmxA -[_ *m (D_rho (i+1)) *m _]mulmxA (nilpotency (C rho) i) mulmx0 !mul0mx.
have ->: (ker_min (row_mx (F_rho (i + 1 + 1)) (H_rho (i + 1 + 1))) *m 
      D_rho (i + 1) *m (D_rho i *m H_rho i *m col_ebase (row_mx 
      (F_rho (i + 1)) (H_rho (i + 1))) *m  castFiHi (i + 1))) = 0.
 by rewrite mulmxA mulmxA mulmxA -[_ *m (D_rho (i+1)) *m _]mulmxA (nilpotency (C rho) i) mulmx0 !mul0mx.
move : (eqmx_trans (ker_intersection (F_rho (i+1+1)) (D_rho (i+1))) (ker_min_kermx (row_mx (F_rho (i+1+1)) (D_rho (i+1))))).
move/eqmxP; move/andP => [_ Hb].
move : Hb; move/submxP => [D' ->].
move : (capmxMr (kermx (F_rho (i+1+1))) (kermx (D_rho (i+1))) (D_rho (i+1))).
move/submxP => [D HD].
rewrite -[D' *m _ *m _]mulmxA HD mulmx_ker capmx0 !mulmx0 !mul0mx.
have -> : (ker_min (row_mx (F_rho (i+1+1)) (H_rho (i+1+1))) *m (D_rho (i+1)) *m (F_rho (i+1))) = 0.
  rewrite -mulmxA  ((M_well_defined (F rho)) (i+1)) mulmxA.
  move : (eqmx_trans (ker_intersection (F_rho (i+1+1)) (H_rho (i+1+1))) (ker_min_kermx (row_mx (F_rho (i+1+1)) (H_rho (i+1+1))))).
  move/eqmxP; move/andP => [_ Hb].
  move : Hb; move/submxP => [D'' ->].
  move : (capmxMr (kermx (F_rho (i+1+1))) (kermx (H_rho (i+1+1))) (F_rho (i+1+1))).
  move/submxP => [p Hp].
  by rewrite -[D'' *m _ *m _]mulmxA Hp mulmx_ker cap0mx !mulmx0 !mul0mx.
by done.
Qed.


(*In this descomposition, two lemmas have to be verified:

  1) D_12 is isomorphic to H_21.
     To this aim, 1.1) D_12 i * H_21 i = Id and
                  1.2) H_21 (i+1) * D_12 (i+1) = Id 
  2) F_13 is isomorphic to G_31.
     F_13 * G_31 = Id  and G_31 * F_13 = Id 
       This is trivial because F_13 and G_31 are identity matrices
*)


(*Auxiliar lemmas to prove 1)  *)
Lemma aux1': forall (i:Z), H_rho_change_base i *m D_rho_change_base i =  block_mx 0 (row_mx 0 0) (col_mx 0 0) (block_mx  (H_aux i *m D_aux i) 0 0 0).
Proof.
move=> i.
by rewrite H_rho_change_baseP D_rho_change_baseP mulmx_block !mul0mx !mulmx0 !GRing.add0r !mul_row_col mul_block_col !row_mx0 !col_mx0 block_mx0 mul_col_row !mul0mx !GRing.add0r col_mx0 GRing.addr0 mulmx0.
Qed.



Lemma aux1 : forall i:Z, Fi_isom (i + 1) *m (H_rho (i + 1 + 1) *m D_rho (i + 1 + 1)) *m   Gi_isom (i + 1) = block_mx 0 (row_mx 0 0) (col_mx 0 0) (block_mx (H_aux (i+1) *m D_aux (i+1)) 0 0 0).
Proof.
move=> i.
have ->:  Fi_isom (i + 1) *m (H_rho (i + 1 + 1) *m D_rho (i + 1 + 1)) *m   Gi_isom (i + 1)  =  Fi_isom (i + 1) *m (H_rho (i + 1 + 1) *m ((Gi_isom (i+1+1)) *m (Fi_isom (i+1+1))) *m D_rho (i + 1 + 1)) *m   Gi_isom (i + 1) by rewrite (GiFi (i+1+1)) mulmx1.
set x:= block_mx 0 (row_mx 0 0) (col_mx 0 0)
     (block_mx (H_aux (i + 1) *m D_aux (i + 1)) 0 0 0).
rewrite !mulmxA.
have ->: ( Fi_isom (i + 1) *m H_rho (i + 1 + 1) *m Gi_isom (i + 1 + 1) *m 
   Fi_isom (i + 1 + 1) *m D_rho (i + 1 + 1) *m Gi_isom (i + 1) = 
  (Fi_isom (i + 1) *m H_rho (i + 1 + 1) *m Gi_isom (i + 1 + 1)) *m 
  ( Fi_isom (i + 1 + 1) *m D_rho (i + 1 + 1) *m Gi_isom (i + 1))) by rewrite !mulmxA.
by rewrite -/(D_rho_change_base (i+1)) -/(H_rho_change_base (i+1)) (aux1' (i+1)).
Qed.



Lemma aux2': forall (i:Z), F_rho_change_base i *m G_rho_change_base i =  block_mx 0 (row_mx 0 0) (col_mx 0 0) (block_mx  0 0 0 1%:M).
move=> i.
by rewrite F_rho_change_baseP G_rho_change_baseP !mul_col_row !row_mx0 !col_mx0 !mul0mx !mulmx0 mulmx1.
Qed.


Lemma aux2 : forall i:Z , Fi_isom (i + 1) *m (F_rho (i + 1 + 1) *m G_rho (i + 1 + 1)) *m  Gi_isom (i + 1) =  block_mx 0 (row_mx 0 0) (col_mx 0 0) (block_mx  0 0 0 1%:M).
Proof.
move=> i.
by rewrite -(aux2' (i+1)) !mulmxA.
Qed.




(* Main lemma *)
(*We prove that the block of D*H is the identity 1.1) *)
Lemma D_rho_i_change_base_H_rho_i_change_base : forall i:Z,
 (ulsubmx ((D_rho_change_base i) *m (H_rho_change_base i))) = 1%:M.
Proof.
move=>i.
rewrite /(D_rho_change_base i) /(H_rho_change_base i) !mulmxA -[_ *m (Gi_isom i) *m _]mulmxA GiFi mulmx1 -[_ *m (D_rho (i+1)) *m _]mulmxA.
have -> : ((D_rho (i+1)) *m (H_rho (i+1))) = 1%:M - ((H_rho (i+1+1)) *m (D_rho (i+1+1))) - ((F_rho (i+1+1)) *m (G_rho (i+1+1))).
  rewrite -(ax2 rho (i+1)) -[(F_rho (i+1+1)) *m (G_rho (i+1+1)) + _ + _]GRing.addrA [(H_rho (i+1+1)) *m _ + _]GRing.addrC GRing.addrA. 
  have -> : (F_rho (i+1+1)) *m (G_rho (i+1+1)) + (D_rho (i+1)) *m (H_rho (i+1)) + (H_rho (i+1+1)) *m (D_rho (i+1+1)) - (H_rho (i+1+1)) *m (D_rho (i+1+1)) - (F_rho (i+1+1)) *m (G_rho (i+1+1)) =
    (F_rho (i+1+1)) *m (G_rho (i+1+1)) + (D_rho (i+1)) *m (H_rho (i+1)) + ((H_rho (i+1+1)) *m (D_rho (i+1+1)) - (H_rho (i+1+1)) *m (D_rho (i+1+1))) - (F_rho (i+1+1)) *m (G_rho (i+1+1)); first by rewrite GRing.addrA. 
  have H4 : (H_rho (i+1+1)) *m (D_rho (i+1+1)) == (H_rho (i+1+1)) *m (D_rho (i+1+1)) by done.
  rewrite -GRing.subr_eq0 in H4.
  rewrite (eqP H4) GRing.addr0 [(_ *m_) + (_ *m _)]GRing.addrC -GRing.addrA; clear H4.
  have H4 : (F_rho (i+1+1)) *m (G_rho (i+1+1)) == (F_rho (i+1+1)) *m (G_rho (i+1+1)) by done.
  rewrite -GRing.subr_eq0 in H4.
  by rewrite (eqP H4) GRing.addr0. 
by rewrite !mulmx_addr !mulmx_addl mulmx1 !mulmxN !mulNmx(FiGi (i+1))aux1 aux2 scalar_mx_block block_mxEv opp_col_mx opp_col_mx !add_col_mx !opp_row_mx !add_row_mx /ulsubmx col_mxKu row_mxKl !GRing.oppr0  !GRing.addr0.
Qed.



(*Namely, the blocks in which we are interested are D_aux and H_aux*)
Lemma D_aux_H_aux :forall i:Z,
 (D_aux i) *m (H_aux i) = 1%:M.
Proof.
move =>i.
move : ( D_rho_i_change_base_H_rho_i_change_base i).
by rewrite D_rho_change_baseP H_rho_change_baseP mulmx_block block_mxKul mul_row_col mul0mx mulmx0 GRing.add0r GRing.addr0.
Qed.


Lemma aux3': forall (i:Z), D_rho_change_base i *m H_rho_change_base i =  block_mx (D_aux i *m H_aux i) (row_mx 0 0) (col_mx 0 0) (block_mx 0 0 0 0).
Proof.
move=> i.
by rewrite H_rho_change_baseP D_rho_change_baseP mulmx_block !mul0mx !mulmx0 !GRing.add0r !mul_row_col mul_block_col !row_mx0 !col_mx0 block_mx0 !mul0mx !GRing.add0r  !GRing.addr0 !mulmx0 col_mx0 block_mx0.
Qed.


(*And the another one: 1.2)  *)
Lemma H_aux_D_aux : forall i:Z,
 (H_aux (i+1)) *m (D_aux (i+1)) = 1%:M.
Proof.
move => i.
have : (H_rho_change_base (i+1)) *m (D_rho_change_base (i+1)) = 1%:M - ((D_rho_change_base (i) *m (H_rho_change_base (i)))) - ((F_rho_change_base (i+1)) *m (G_rho_change_base (i+1))).
  rewrite -(gf_Dh_hD_change_base i). 
  have ->: F_rho_change_base (i + 1) *m G_rho_change_base (i + 1) +
   H_rho_change_base (i + 1) *m D_rho_change_base (i + 1) +
   D_rho_change_base i *m H_rho_change_base i -
   D_rho_change_base i *m H_rho_change_base i -
   F_rho_change_base (i + 1) *m G_rho_change_base (i + 1) = 
 F_rho_change_base (i + 1) *m G_rho_change_base (i + 1) +
   H_rho_change_base (i + 1) *m D_rho_change_base (i + 1) +
   (D_rho_change_base i *m H_rho_change_base i -
   D_rho_change_base i *m H_rho_change_base i) -
   F_rho_change_base (i + 1) *m G_rho_change_base (i + 1); first by rewrite GRing.addrA.
  have H4 : (D_rho_change_base i *m H_rho_change_base i) == (D_rho_change_base i *m H_rho_change_base i) by done.
  rewrite -GRing.subr_eq0 in H4.
  rewrite (eqP H4) GRing.addr0 -[(F_rho_change_base (i+1)) *m (G_rho_change_base (i+1)) + _ + _]GRing.addrA [(H_rho_change_base (i+1)) *m _ + _]GRing.addrC  GRing.addrA; clear H4.
  have H4 : F_rho_change_base (i + 1) *m G_rho_change_base (i + 1)== F_rho_change_base (i + 1) *m G_rho_change_base (i + 1) by done.
  rewrite -GRing.subr_eq0 in H4.
  by rewrite (eqP H4) GRing.add0r.
rewrite /F_rho_change_base /G_rho_change_base  H_rho_change_baseP D_rho_change_baseP mulmx_block !row_mx0 !col_mx0 !mulmx0 !GRing.add0r /D_rho_change_base /H_rho_change_base !mul0mx block_mx0 mul_col_mx !mul0mx !GRing.add0r GRing.addr0 mul_mx_row mulmx0.
have -> : Fi_isom (i + 1) *m F_rho (i + 1 + 1) *m (G_rho (i + 1 + 1) *m  Gi_isom (i + 1))= (Fi_isom (i + 1) *m (F_rho (i + 1 + 1) *m  G_rho(i + 1 + 1)) *m Gi_isom (i + 1)); first by rewrite !mulmxA.
rewrite (aux2 i) (aux3' i) !opp_block_mx -GRing.addrA add_block_mx scalar_mx_block  add_block_mx add_block_mx =>  H.
move : (eq_block_mx H) => H1; case : H1. 
move => H1 H2 H3; rewrite scalar_mx_block add_block_mx block_mxEv => H4.
move: (eq_col_mx H4); move => H5; case H5 => H6 H7.
move: (eq_row_mx H6) => H8; case H8.
by rewrite !GRing.subr0. 
Qed.

End new_D_new_H.

End isomorphism.

End definitions.

End decomposition.


