Add LoadPath "CoqEAL".

Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype.
Require Import finfun tuple ssralg matrix mxalgebra bigop.
Require Import general_decomposition ZArith Zinfra definitions_auxlemmas.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


(*Hexagonal Lemma*)

Section hexagonal_lemma.

Local Open Scope ring_scope.

(*We start with a chain complex CH where every differential is a matrix composed by
3 x 3 blocks. *)

Variable K : fieldType.
Variables (m1 m2 m3 : Z -> nat).
Variable Di : forall i:Z, 'M[K]_((m1 (i+1))+((m2 (i+1))+(m3 (i+1))),((m1 i)+ ((m2 i) + (m3 i)))).
Hypothesis nilpotency : forall i:Z, Di (i+1) *m Di i = 0.

Definition CH := Build_FGChain_Complex nilpotency.


Section descomposition_D1D2.

(*Every differential is divided into 9 blocks*)
Definition daa (i:Z) := (ulsubmx (diff CH i)).
Definition dab (i:Z) := (lsubmx (ursubmx  (diff CH i))).
Definition dac (i:Z) := (rsubmx (ursubmx (diff CH i))).
Definition dba (i:Z) := usubmx (dlsubmx (diff CH i)).
Definition dbb (i:Z) := (ulsubmx (drsubmx  (diff CH i))).
Definition dbc (i:Z) := (ursubmx (drsubmx  (diff CH i))).
Definition dca  (i:Z):= dsubmx(dlsubmx (diff CH i)).
Definition dcb (i:Z) := (dlsubmx (drsubmx  (diff CH i))).
Definition dcc (i:Z) := (drsubmx (drsubmx  (diff CH i))).

Lemma decomposition_D1 : forall i:Z, (diff CH i) =
  block_mx (daa i) (row_mx (dab i) (dac i)) (col_mx (dba i) (dca i)) (block_mx (dbb i) (dbc i) (dcb i) (dcc i)).
Proof.
move=> i.
by rewrite -(submxK (diff CH i)) (hsubmxK (ursubmx (diff CH i))) (vsubmxK (dlsubmx (diff CH i)))
 (submxK (drsubmx (diff CH i))).
Qed.



Section auxiliary_lemmas.

(* The following auxiliar lemmas are obtained from the property of the chain complex:
   Di (i+1) * Di i = 0.

They are used in the moment of proving that exists a reduction of the descomposed CH*)

Lemma D1D2_1 : forall i:Z,
  daa (i+1) *m daa i + dab (i+1) *m dba i + dac (i+1) *m dca i = 0.
Proof.
move=> i.
move: (nilpotency i).
move: (decomposition_D1 i)(decomposition_D1 (i+1)).
rewrite /= => -> ->.
rewrite  mulmx_block !mul_row_col 
        -block_mx0 !GRing.addrA => H1.
by move : (eq_block_mx H1) => [-> _].
Qed. 


Lemma D1D2_2 : forall i:Z,
   daa (i+1) *m dac i + dab (i+1) *m dbc i + dac (i+1) *m dcc i = 0.
Proof.
move=> i.
move: (nilpotency i).
move: (decomposition_D1 i)(decomposition_D1 (i+1)).
rewrite /=  => -> ->.
rewrite  mulmx_block !mul_row_col
        -block_mx0 !GRing.addrA.
rewrite !mul_block_col !mul_mx_row !add_row_mx => H1.
move : (eq_block_mx H1) => [H2 H3 H4 H5].
move: H3; rewrite -row_mx0 =>H3.
by move: (eq_row_mx H3) => [H6 HH].
Qed.

Lemma D1D2_21 : forall i:Z,
   daa (i+1) *m dab i + dab (i+1) *m dbb i + dac (i+1) *m dcb i = 0.
Proof.
move=> i.
move: (nilpotency i).
move: (decomposition_D1 i)(decomposition_D1 (i+1)).
rewrite /=  => -> ->.
rewrite  mulmx_block !mul_row_col
        -block_mx0 !GRing.addrA.
rewrite !mul_block_col !mul_mx_row !add_row_mx => H1.
move : (eq_block_mx H1) => [H2 H3 H4 H5].
move: H3; rewrite -row_mx0 =>H3.
by move: (eq_row_mx H3) => [H6 HH].
Qed.



Lemma D1D2_3 : forall i:Z,
   dca (i+1) *m daa i + dcb (i+1) *m dba i + dcc (i+1) *m dca i = 0.
Proof.
move=> i.
move: (nilpotency i).
move: (decomposition_D1 i)(decomposition_D1 (i+1)).
rewrite /= => -> ->.
rewrite  mulmx_block !mul_row_col 
        -block_mx0 !GRing.addrA.
rewrite !mul_block_col mul_col_mx add_col_mx => H1.
move : (eq_block_mx H1) => [H2 H3 H4 H5].
move: H4; rewrite -col_mx0 =>H4.
move: (eq_col_mx H4) => [_ HH].
by move: HH; rewrite GRing.addrA => ->.
Qed. 


Lemma D1D2_32 : forall i:Z,
   dca (i+1) *m dab i + dcb (i+1) *m dbb i + dcc (i+1) *m dcb i = 0.
Proof.
move=> i.
move: (nilpotency i).
move: (decomposition_D1 i)(decomposition_D1 (i+1)).
rewrite /= => -> ->.
rewrite  mulmx_block !mul_row_col 
        -block_mx0 !GRing.addrA !mul_block_col mul_col_mx add_col_mx => H1.
move : (eq_block_mx H1) => [H2 H3 H4 H5].
move: H5; rewrite mul_col_row.
rewrite block_mxEv add_col_mx -col_mx0 =>H41.
move: (eq_col_mx H41) => [_ HH].
move: HH; rewrite -row_mx0 !mul_mx_row !add_row_mx => H5.
move: (eq_row_mx H5) =>[h6 h7].
by move: h6; rewrite GRing.addrA => ->.
Qed. 



Lemma D1D2_31 : forall i:Z,
   dca (i+1) *m dab i + dcb (i+1) *m dbb i + dcc (i+1) *m dcb i = 0.
Proof.
move=> i.
move: (nilpotency i).
move: (decomposition_D1 i)(decomposition_D1 (i+1)).
rewrite /= => -> ->.
rewrite  mulmx_block !mul_row_col  -block_mx0 !GRing.addrA !mul_block_col mul_col_mx
  !add_col_mx mul_col_row => H1.
move : (eq_block_mx H1) => [H2 H3 H4 H5].
move:H5; rewrite block_mxEv add_col_mx. 
rewrite -col_mx0 =>H5.
move: (eq_col_mx H5) => [ HH h].
move: h; rewrite -row_mx0 !mul_mx_row !add_row_mx => H6. 
move: (eq_row_mx H6) => [h7 h8].
by move: h7; rewrite GRing.addrA => ->.
Qed. 



Lemma D1D2_4 : forall i:Z, (dca (i+1) *m dac i + dcb (i+1) *m dbc i + dcc (i+1) *m dcc i) = 0.
Proof.
move=> i.
move: (nilpotency i).
move: (decomposition_D1 i)(decomposition_D1 (i+1)).
rewrite /= => -> ->.
rewrite  mulmx_block !mul_row_col  -block_mx0 !GRing.addrA !mul_block_col mul_col_mx 
   !add_col_mx !mul_mx_row !add_row_mx => H1.
move : (eq_block_mx H1) => [_ _ _ H5].
move: H5; rewrite -block_mx0.
rewrite -block_mxEv !mul_col_mx -block_mxEh add_block_mx=> H4.
move: (eq_block_mx H4) => [_ _ _ H2].
by move: H2; rewrite GRing.addrA => ->.
Qed. 




Lemma m_m_0 (R : ringType) (t1 t2 : nat) (M : 'M[R]_(t1,t2)) : M - M = 0.
Proof.
by move : (eq_refl M); rewrite -GRing.subr_eq0; move/eqP => ->.
Qed.


Lemma d'ccdca : forall i:Z, 
  (dcc (i+1)) *m (dcb (i)) =  - ((dcb (i+1)) *m (dbb i)) - (dca (i+1)) *m (dab i).
Proof.
move=> i.
have H1: -  dca(i+1) *m dab i +  dcb (i+1) *m dbb i +  dca(i+1) *m dab i +
   dcc(i+1) *m dcb i =  -  dca(i+1) *m dab i.
rewrite -!GRing.addrA [dcb (i+1) *m dbb i + (_ + _)]GRing.addrA.  
  move: (D1D2_31 i). 
  by rewrite [dca (i+1) *m _ + _]GRing.addrC=> ->; rewrite GRing.addr0.
have H2 : 
   -  dcb(i+1) *m dbb i - dca(i+1) *m dab i + dcb(i+1) *m dbb i + dca(i+1) *m dab i + 
   dcc(i+1) *m dcb i =
   - dcb(i+1) *m dbb i - dca(i+1) *m dab i.
rewrite -{2}mulNmx.
by rewrite -H1 !GRing.addrA -mulNmx.
move : H2.
rewrite -!mulNmx [- (dcb (i+1)) *m (dbb i) + _]GRing.addrC -!GRing.addrA
  [-(dcb (i+1)) *m (dbb i) + (_ + _)]GRing.addrA
  [- (dcb (i+1)) *m (dbb i) + _]GRing.addrC !mulNmx.
move =>  <-.
by rewrite m_m_0 GRing.add0r GRing.addrA -mulmxN [dca (i+1) *m _ + _]GRing.addrC mulmxN m_m_0 GRing.add0r. 
Qed.




Lemma d'acdcc : forall i:Z,
   dac (i+1) *m dcc i = - daa (i+1) *m dac(i) - dab (i+1) *m dbc i.
Proof.
move=> i.
have H1 : -  dab (i+1) *m dbc i +daa (i+1) *m dac(i) +  dab (i+1) *m dbc i +
    dac (i+1) *m dcc i = - dab (i+1) *m dbc i; first by rewrite -!GRing.addrA [daa (i+1) *m dac i + (_ + _)]GRing.addrA D1D2_2  GRing.addr0.
have H2 :  - daa (i+1) *m dac i - dab(i+1) *m dbc i + daa(i+1) *m dac i +
    dab(i+1) *m dbc i + dac(i+1) *m dcc i=
  - daa (i+1) *m dac i - dab(i+1) *m dbc i.
by rewrite -!GRing.addrA [daa(i+1) *m dac i + (_ + _)]GRing.addrA
  ![- (dab (i+1) *m dbc i) + (_ + _)]GRing.addrA -mulNmx H1.
move : H2.
rewrite -!mulNmx [- daa(i+1) *m dac i + _]GRing.addrC -!GRing.addrA
  [- daa(i+1) *m dac i + (_ + _)]GRing.addrA
  [- daa(i+1) *m dac i + _]GRing.addrC !mulNmx.
move : (eq_refl (daa(i+1) *m dac i)).
rewrite -GRing.subr_eq0; move/eqP => ->; rewrite GRing.add0r -mulNmx GRing.addrA [- dab(i+1) *m _ + _]GRing.addrC mulNmx.
move : (eq_refl (dab(i+1) *m dbc i)).
by rewrite -GRing.subr_eq0; move/eqP => ->; rewrite GRing.add0r.   
Qed.


Lemma d'cbdbc : forall i:Z, 
  dcb (i+1) *m dbc i = - dca (i+1) *m dac i - dcc (i+1) *m dcc i.
Proof.
move=> i.
move : D1D2_1 D1D2_2 D1D2_3 D1D2_4 => D1D2_1 D1D2_2 D1D2_3 D1D2_4.
move : (D1D2_4 i).
rewrite -GRing.addrA [dcb (i+1) *m _ + _]GRing.addrC GRing.addrA => D1D2_41.
have H1 : - dcc(i+1) *m dcc i + dca (i+1) *m dac i + dcc(i+1) *m dcc i + 
   dcb(i+1) *m dbc i = - dcc(i+1) *m dcc i; first by rewrite -!GRing.addrA [dca(i+1) *m dac i + (_ + _)]GRing.addrA D1D2_41 GRing.addr0.
have H2 : - dca(i+1) *m dac i - dcc(i+1) *m dcc i + dca(i+1) *m dac i + 
  dcc(i+1) *m dcc i +  dcb(i+1) *m dbc i = - dca(i+1) *m dac i - dcc(i+1) *m dcc i;   first by rewrite -!GRing.addrA [dca(i+1) *m dac i + (_ + _)]GRing.addrA 
   ![- (dcc(i+1) *m dcc i) + (_ + _)]GRing.addrA -mulNmx H1.
move : H2.
rewrite -!mulNmx [- dca(i+1) *m dac i + _]GRing.addrC -!GRing.addrA
  [- dca(i+1) *m dac i + (_ + _)]GRing.addrA
  [- dca(i+1) *m dac i + _]GRing.addrC !mulNmx.
move : (eq_refl (dca(i+1) *m dac i)).
rewrite -GRing.subr_eq0; move/eqP => ->; rewrite GRing.add0r -mulNmx GRing.addrA [- dcc(i+1) *m _ + _]GRing.addrC mulNmx.
move : (eq_refl (dcc(i+1) *m dcc i)).
by rewrite -GRing.subr_eq0; move/eqP => ->; rewrite GRing.add0r.   
Qed.


Lemma d'acdca :forall i:Z,
  (dac (i+1)) *m dcb i = - (dab (i+1)) *m (dbb i) - (daa (i+1)) *m (dab i).
Proof.
move=> i.
Proof.
move : D1D2_1 D1D2_2 D1D2_3 D1D2_4 => D1D2_1 D1D2_2 D1D2_3 D1D2_4.
have H1 : - dab(i+1) *m dbb i +  daa(i+1) *m dab i + dab(i+1) *m dbb i +
  dac(i+1) *m dcb i  = - dab(i+1) *m dbb i; first by rewrite -!GRing.addrA 
  [daa(i+1) *m dab i + (_ + _)]GRing.addrA D1D2_21 GRing.addr0.
have H2 : - daa(i+1) *m dab i - dab(i+1) *m dbb i + daa(i+1) *m dab i +
   dab(i+1) *m dbb i + dac(i+1) *m dcb i = - daa(i+1) *m dab i - dab(i+1) *m dbb i; 
  first by rewrite -!GRing.addrA [daa(i+1) *m dab i + (_ + _)]GRing.addrA
  ![- (dab(i+1) *m dbb i) + (_ + _)]GRing.addrA -mulNmx H1.
move : H2.
rewrite -!mulNmx [- daa(i+1) *m dab i + _]GRing.addrC -!GRing.addrA
  [- daa(i+1) *m dab i + (_ + _)]GRing.addrA
  [- daa(i+1) *m dab i + _]GRing.addrC !mulNmx.
move : (eq_refl (daa(i+1) *m dab i)).
rewrite -GRing.subr_eq0; move/eqP => ->; rewrite GRing.add0r -mulNmx GRing.addrA [- dab(i+1) *m _ + _]GRing.addrC mulNmx.
move : (eq_refl (dab(i+1) *m dbb i)).
by rewrite -GRing.subr_eq0; move/eqP => ->; rewrite GRing.add0r.   
Qed.




Lemma d'cadaa : forall i:Z, dca(i+1) *m dab i = - dcc(i+1) *m dcb i - dcb (i+1) *m dbb i.
Proof.
move=> i.
move : D1D2_1 D1D2_2 D1D2_3 D1D2_4 => D1D2_1 D1D2_2 D1D2_3 D1D2_4.
move : (D1D2_32 i).
rewrite -GRing.addrA [dcb(i+1) *m _ + _]GRing.addrC
        [dca(i+1) *m _ + _]GRing.addrC => D1D2_31.
have H1 : - dcb(i+1) *m dbb i + dcc(i+1) *m dcb i + dcb(i+1) *m dbb i 
   + dca(i+1) *m dab i =  - dcb(i+1) *m dbb i; first by rewrite -!GRing.addrA
   [dcc(i+1) *m dcb i + (_ + _)]GRing.addrA D1D2_31 GRing.addr0.
have H2 : - dcc(i+1) *m dcb i - dcb(i+1) *m dbb i + dcc(i+1) *m dcb i + dcb(i+1) *m dbb i +  dca(i+1) *m dab i = - dcc(i+1) *m dcb i - dcb(i+1) *m dbb i.
by rewrite -!GRing.addrA [dcc(i+1) *m dcb i + (_ + _)]GRing.addrA
  ![- (dcb(i+1) *m dbb i) + (_ + _)]GRing.addrA -mulNmx H1.
move : H2.
rewrite -!mulNmx [- dcc(i+1) *m dcb i + _]GRing.addrC -!GRing.addrA
  [- dcc(i+1) *m dcb i + (_ + _)]GRing.addrA
  [- dcc(i+1) *m dcb i + _]GRing.addrC !mulNmx.
move : (eq_refl (dcc(i+1) *m dcb i)).
rewrite -GRing.subr_eq0; move/eqP => ->; rewrite GRing.add0r -mulNmx GRing.addrA [- dcb(i+1) *m _ + _]GRing.addrC mulNmx.
move : (eq_refl (dcb(i+1) *m dbb i)).
by rewrite -GRing.subr_eq0; move/eqP => ->; rewrite GRing.add0r.  
Qed.


Lemma subr_sub (R : ringType) (t1 t2 : nat) (m n p : 'M[R]_(t1,t2)) : m - n - p = m - (n + p).
Proof.
apply/eqP.
by rewrite !GRing.subr_eq -GRing.addrA [p+n]GRing.addrC GRing.subrK.
Qed.


Lemma subr_opp (R : ringType) (t1 t2 : nat) (m n : 'M[R]_(t1,t2)) : 
  m - - n = m + n.
Proof.
apply/eqP.
rewrite !GRing.subr_eq -GRing.addrA. 
move : (eq_refl n) => H.
rewrite -GRing.subr_eq0 in H.
by rewrite (eqP H) GRing.addr0.
Qed.

Lemma subr_add (R : ringType) (t1 t2 : nat) (m n p : 'M[R]_(t1,t2)) : 
 m - (n - p) = m - n + p.
Proof.
apply/eqP.
rewrite !GRing.subr_eq -!GRing.addrA [n-p]GRing.addrC [p + _]GRing.addrA.
move : (eq_refl p) => H.
rewrite -GRing.subr_eq0 in H.
rewrite (eqP H) GRing.add0r [_ + n]GRing.addrC.
move : (eq_refl n) => H1.
rewrite -GRing.subr_eq0 in H1.
by rewrite (eqP H1) GRing.addr0.
Qed.        

End auxiliary_lemmas.


End descomposition_D1D2.

(*End auxiliary_lemmas.*)





Section definitions.

(*Apart from the chain complex CH where every differential is a 3 x 3 matrix, we have
  as hypothesis that dab (whose dimension is (m1 (i+1), m2 i)) is a isomorphism. Then,  we define a variable dab_1 satisfying dba_invertible. *)

Variable dab_1 : forall i : Z, 'M[K]_(m2 i, m1 (i + 1)).

Hypothesis dba_invertible : 
  forall i:Z, (dab (i+1)) *m (dab_1 (i+1)) = 1%:M /\ (dab_1 (i+1)) *m (dab (i+1)) = 1%:M.


Lemma cast_1_minus1' :forall i:Z, (i - 1 + 1)%Z = i.
Proof.
move =>i.
by rewrite subr_addr.
Qed.

Lemma dba_invertiblei : 
forall i:Z, (dab (i)) *m (dab_1 (i)) = 1%:M /\ (dab_1 (i)) *m (dab (i)) = 1%:M.
Proof.
move => i.
have: ((dab (i-1+1)) *m (dab_1 (i-1+1))) = 1%:M /\ (dab_1 (i-1+1)) *m (dab (i-1+1)) = 1%:M by exact :(dba_invertible (i-1)).
move=> H1.
have <-: i - 1 + 1  = i; first by exact : cast_1_minus1'.
exact: H1.
Qed. 



(*With these hypotheses, we can build a reduction between the chain complex CH

 To this aim, we define di, the morphisms fi and gi and homotopy operator hi.

 Let us note that to define (gi) uses blocks which are in dimension i and (i-1).
 So, we have to define (gi i) as a matrix (i+1) -> (i+1) instead of i -> i.
 Then, fi and di have had to be define according to this restriction.

 Moreover, the reduced chain complex which have just to define with di is moved
   one dimension from CH.

*)
  

Definition di_prev (i:Z):= (dcc i) - ((dcb i) *m (dab_1 i)) *m (dac i).
Definition di (i:Z):= di_prev (i+1).

Definition fi (i:Z) := (col_mx 0 (col_mx (- (dab_1 (i+1)) *m (dac (i+1))) 1%:M))
   :'M[K]_((m1 (i+1))+((m2 (i+1))+(m3 (i+1))),m3 (i+1)).

Definition gi (i:Z) := 
  (row_mx (- (dcb i) *m (dab_1 i)) (row_mx 0 1%:M)) : 'M[K]_(m3 (i+1),(m1 (i+1))+((m2 (i+1))+(m3 (i+1)))).

Definition hi (i:Z) := 
  (block_mx 0 (row_mx 0 0) (col_mx (dab_1 i) 0) 0): 'M[K]_((m1 i)+ ((m2 i) + (m3 i)), ((m1 (i+1))+((m2 (i+1))+(m3 (i+1))))).




(*We prove the property so that the chain complex which is composed by the 
  differential di_previ is a chain complex*)
Lemma ddi_prev  : forall i:Z,  di_prev (i+1) *m di_prev i = 0.
Proof.
move => i.
move : D1D2_1 D1D2_2 D1D2_3 D1D2_4 => D1D2_1 D1D2_2 D1D2_3 D1D2_4.
rewrite !mulmx_addl !mulmx_addr !mulmxN !mulNmx !GRing.addrA -!mulmxA
        d'acdcc !mulmx_addr !mulNmx !mulmxN.
have -> : dcb (i+1) *m (dab_1 (i+1) *m (dab (i+1) *m dbc i)) = dcb (i+1) *m dbc i.
  rewrite [dab_1 (i+1) *m _]mulmxA.
  move : (dba_invertible i) => [ _ ->].
  by rewrite mul1mx.
rewrite -!mulNmx -subr_sub !mulNmx !subr_opp d'cbdbc !GRing.addrA [dcc (i+1) *m dcc i + _]GRing.addrC [(- (dcc (i + 1) *m (dcb i *m (dab_1 i *m dac i))) + dcc (i + 1) *m dcc i) + _]GRing.addrC [ dcb (i + 1) *m (dab_1 (i + 1) *m (daa (i + 1) *m dac i)) +
   (- (dcc (i + 1) *m (dcb i *m (dab_1 i *m dac i))) + dcc (i + 1) *m dcc i) + _]GRing.addrC -!GRing.addrA [dcc (i+1) *m dcc i + _]GRing.addrA.
move : (eq_refl (dcc (i+1) *m dcc i)).
rewrite -GRing.subr_eq0; move/eqP => ->.
rewrite GRing.add0r [dac (i+1) *m _]mulmxA d'acdca mulmx_addl !mulmx_addr.
have -> :  dcb (i + 1) *m (dab_1 (i + 1) *m (- (daa (i + 1) *m dab i) *m 
  (dab_1 i *m  dac i))) = - dcb (i + 1) *m (dab_1 (i + 1) *m (daa (i + 1) *m dac i)).
  rewrite -mulNmx -!mulmxA [dab i *m _]mulmxA.
  move : (dba_invertiblei i) => [-> _ ].
  by rewrite mul1mx !mulNmx !mulmxN. 
have -> : (dcb (i + 1) *m (dab_1 (i + 1) *m (- dab (i + 1) *m  dbb i *m (dab_1 i *m 
  dac i))) = - dcb (i + 1) *m dbb i *m dab_1 i *m dac i).
  rewrite mulNmx -!mulmxA [dab_1(i+1) *m _]mulmxA mulmxN [dab_1(i+1) *m _]mulmxA.
  move : (dba_invertible i)=> [_ ->].
  by rewrite mul1mx !mulNmx !mulmxN !mulmxA.
rewrite !GRing.addrA [ - dca (i+1) *m dac i + _]GRing.addrC [_ + - dcb (i+1) *m (dab_1  (i+1) *m (daa(i+1) *m dac i))]GRing.addrC !GRing.addrA [- dcb(i+1) *m (dab_1(i+1) *m (daa(i+1) *m dac i)) + _]GRing.addrC mulNmx.
move : (eq_refl (dcb (i+1) *m (dab_1(i+1) *m (daa(i+1) *m dac i)))).
rewrite -GRing.subr_eq0; move/eqP => ->.
rewrite GRing.add0r -!mulmxA !mulNmx -GRing.addrA.
have -> : (- (dcc (i + 1) *m (dcb i *m (dab_1 i *m dac i))) -
    dcb (i + 1) *m (dbb i *m (dab_1 i *m dac i))) =
    (- (dcc (i + 1) *m dcb i) - (dcb (i + 1) *m dbb i)) *m (dab_1 i *m dac i)
  by rewrite !mulmx_addl !mulNmx !mulmxA.   
have -> : (- (dcc(i+1) *m dcb i) - dcb (i+1) *m dbb i) = dca(i+1) *m dab i by rewrite -mulNmx -d'cadaa.
rewrite -mulmxA [dab i *m _]mulmxA.
move : (dba_invertiblei i) => [-> _].
rewrite mul1mx GRing.addrC.
by move : (eq_refl (dca (i+1) *m dac i)); rewrite -GRing.subr_eq0; move/eqP => ->.
Qed.


Lemma ddi : forall i:Z,  di (i+1) *m di i = 0.
Proof. 
move => i.
exact: ddi_prev (i+1).
Qed.

Definition m3'(i:Z):=m3(i+1).

(*We define the new chain complex newD' composed by di*)
Definition newD' := Build_FGChain_Complex (m:=m3')(diff:= di) ddi.



(*Up to now, we have a chain complex initial CH and another one reduced newD'
  where have been also defined the fi, gi, hi and di but we cannot prove that
  exists a reduction between both chain complexes because newD' is moved one 
  dimension.

  Then, we are going to build a new chain complex from CH which is moved one dimension  which we call Di_elevado*)

Section CC_elevado.


Definition diffi_elevado(i:Z):= (diff CH (i+1)). 

Lemma diffi_elevado_nilpotency: forall i:Z, diffi_elevado(i+1)*m diffi_elevado(i)= 0.
Proof.
move=> i.
exact : (definitions_auxlemmas.nilpotency CH (i+1)).
Qed.

Definition Di_elevado:= Build_FGChain_Complex diffi_elevado_nilpotency.


(*With this new definition of the differential, we can prove the properties necessary to define the morphisms fe and ge*)


(*(diff CH (i+1)) = diff Di_elevado i*)
Lemma df_fDi : forall i:Z, (diff Di_elevado i) *m (fi i)= (fi (i+1)) *m (di i).
Proof. 
move => i.
have ->: (diff Di_elevado i) = diff CH (i+1) by done.
rewrite decomposition_D1  mul_col_mx !mul_col_mx mul0mx mul1mx /fi /di /di_prev !mul_row_col !mulmx0 !mulmx1 !GRing.add0r.
have -> : (dab (i+1) *m (- dab_1 (i+1) *m dac (i+1)) + dac(i+1)) = 0.
  rewrite mulmxA mulmxN.
  move : (dba_invertible i)=> [-> _].
  by rewrite mulNmx mul1mx GRing.addrC m_m_0.
rewrite block_mxEh mul_row_col !mul_col_mx add_col_mx !mulmx1.
have ->:  (dcb (i + 1) *m (- dab_1 (i + 1) *m dac (i + 1)) + dcc (i + 1)) =
    (dcc (i + 1) - dcb (i + 1) *m dab_1 (i + 1) *m dac (i + 1)) by rewrite GRing.addrC mulmxA -mulNmx mulmxN.
have ->: (dbb (i + 1) *m (- dab_1 (i + 1) *m dac (i + 1)) + dbc (i + 1)) =
    (- dab_1 (i + 1 + 1) *m dac (i + 1 + 1) *m 
         (dcc (i + 1) - dcb (i + 1) *m dab_1 (i + 1) *m dac (i + 1))); last by done.
rewrite !mulmxA mulmxN mulmx_addr !mulNmx !mulmxN GRing.addrC subr_opp
        GRing.addrC !mulmxA [_ + dbc (i+1)]GRing.addrC.
have->: (dab_1 (i + 1 + 1) *m dac (i + 1 + 1) *m dcc (i + 1)) =
       (dab_1 (i + 1 + 1) *m (dac (i + 1 + 1) *m dcc (i + 1))) by rewrite !mulmxA.
rewrite d'acdcc !mulmx_addr !mulNmx !mulmxN !mulmxA.
move : (dba_invertible (i+1)) => [ _ ->]. 
rewrite mul1mx -!mulmxA [dac (i+1+1) *m _]mulmxA d'acdca mulmx_addl !mulNmx mulmx_addr
        !mulmxN !mulmxA.
move: (dba_invertible (i+1)) => [_ ->]; rewrite mul1mx.
have ->:  dab_1 (i + 1 + 1) *m daa (i + 1 + 1) *m dab (i + 1) *m 
    dab_1 (i + 1) *m dac (i + 1) =  dab_1 (i + 1 + 1) *m daa (i + 1 + 1)
     *m dac (i + 1).
  rewrite -!mulmxA [(daa (i+1+1)) *m _]mulmxA -mulmxA [(dab (i+1)) *m _]mulmxA.
  by move : (dba_invertible i) => [-> _]; rewrite mul1mx mulmxA.
rewrite -!mulmxN.
have ->: ((dbb (i + 1) *m dab_1 (i + 1) *m - dac (i + 1) +
    dab_1 (i + 1 + 1) *m daa (i + 1 + 1) *m - dac (i + 1)) =
    ( dab_1 (i + 1 + 1) *m daa (i + 1 + 1) *m - dac (i + 1) +
    dbb (i + 1) *m dab_1 (i + 1) *m - dac (i + 1))) by rewrite GRing.addrC.
rewrite -!mulmxA GRing.oppr_sub -mulmxN -GRing.addrA.
have ->: (dab_1 (i + 1 + 1) *m - (daa (i + 1 + 1) *m - dac (i + 1)) +
    (dab_1 (i + 1 + 1) *m (daa (i + 1 + 1) *m - dac (i + 1)) +
     dbb (i + 1) *m (dab_1 (i + 1) *m - dac (i + 1)))) = 
   (dab_1 (i + 1 + 1) *m - (daa (i + 1 + 1) *m - dac (i + 1)) +
    (dab_1 (i + 1 + 1) *m (daa (i + 1 + 1) *m - dac (i + 1)))) +
     dbb (i + 1) *m (dab_1 (i + 1) *m - dac (i + 1))  by rewrite !GRing.addrA.
by rewrite [dab_1 (i + 1 + 1) *m - (daa (i + 1 + 1) *m - dac (i + 1)) + _]GRing.addrC !mulmxN subr_opp -!mulmxN [dab_1 (i + 1 + 1) *m (daa (i + 1 + 1) *m - dac (i + 1)) + _] GRing.addrC !mulmxN m_m_0 GRing.sub0r.
Qed.


Lemma Dg_gdi : forall i:Z, (di i) *m (gi i)=(gi (i+1)) *m (diff Di_elevado i).
Proof.
move => i.
have ->: (diff Di_elevado i) = diff CH (i+1) by done.
rewrite decomposition_D1 mul_row_block !mul_mx_row !mulmx0 !mulmx1 !mul_row_col !mul0mx !mul1mx !GRing.add0r !mulNmx -mulmxA -mulmxA.
elim : (dba_invertiblei i) => H1 H2.
elim : (dba_invertible (i))=>[H3 H4]. 
rewrite H4 mulmx1 GRing.addrC add_row_mx /di /di_prev [- dcb (i+1) + dcb(i+1)]GRing.addrC m_m_0 [_ + dcc (i+1)]GRing.addrC.
have ->:(((dcc (i + 1) - dcb (i + 1) *m dab_1 (i + 1) *m dac (i + 1)) *m 
      - (dcb i *m dab_1 i)) = 
       (dca (i + 1) - dcb (i + 1) *m (dab_1 (i + 1) *m daa (i + 1)))); last by done.
rewrite mulmx_subl mulmxN  mulmxA d'ccdca mulmx_addl !mulNmx.
have -> : (dca(i+1) *m (dab i) *m (dab_1 i)) = dca (i+1) by rewrite -mulmxA H1 mulmx1.
rewrite mulmxN -!mulmxA [(dac (i+1)) *m _]mulmxA d'acdca mulmx_addl mulNmx !mulmx_addr
        !mulmxN -!mulmxA H1 mulmx1 -!mulmxN GRing.oppr_sub [dab_1 (i+1) *m _]mulmxA !mulmxN H4 !mulNmx mul1mx GRing.oppr_sub !mulmxN subr_add subr_opp -mulmxN GRing.addrC.
have ->: dca (i + 1) + dcb (i + 1) *m (dbb i *m dab_1 i) = 
        dcb (i + 1) *m (dbb i *m dab_1 i) + dca (i+1) by rewrite GRing.addrC.
rewrite !GRing.addrA.
have ->: - (dcb (i + 1) *m (dbb i *m dab_1 i)) + dcb (i + 1) *m (dbb i *m dab_1 i) = 0  by rewrite -mulNmx GRing.addrC mulNmx m_m_0.
by rewrite GRing.add0r.
Qed.


Definition fe := Build_FGChain_Complex_Morphism (A:=Di_elevado) (B:=newD') df_fDi.
Definition ge := Build_FGChain_Complex_Morphism (A:=newD') (B:=Di_elevado) Dg_gdi.

(*The homotopy operator is also moved one dimension*)
Definition hie(i:Z):= hi (i+1).
Definition he:= Build_FGHomotopy_operator (A:=Di_elevado) hie.


(*We prove the properties of a reduction between Di_elevado (CH moved one dimension)
  and the chain complex newD' composed by the differentials di*)

Lemma fgi : forall i:Z, (gi i) *m (fi i) = 1%:M.
Proof.
move => i.
by rewrite !mul_row_col mul0mx mulmx0 mul1mx !GRing.add0r.
Qed.


Lemma fhie : forall i:Z, (hi (i+1)) *m (fi (i+1)) = 0.
Proof.
move=> i.
by rewrite mul_block_col !mul_row_col !mulmx0 !mul0mx !GRing.addr0 col_mx0.
Qed.


Lemma hgie : forall i:Z, (gi (i)) *m (hi (i+1)) = 0.
Proof.
move=>i.
by rewrite mul_row_block row_mx0 mul_row_col !mulmx0 !GRing.addr0  mul0mx
        GRing.add0r row_mx0.
Qed.

Lemma hhie : forall i:Z, (hi (i+1)) *m (hi (i+1+1)) = 0.
Proof.
move=>i.
by rewrite mulmx_block !row_mx0 !mul0mx !mulmx0 !GRing.addr0 
   block_mx0.
Qed.


Lemma gf_Dh_hDie : forall i:Z, 
   (fi (i+1)) *m (gi (i+1)) + (hi (i+1+1)) *m (diff CH (i+1+1)) + (diff CH (i+1)) *m (hi (i+1)) = 1%:M.
Proof. 
move=> i.
rewrite (decomposition_D1 (i+1)) decomposition_D1 !mul_col_row !mulmx_block
        !add_block_mx !mulmx1 !mul0mx !mulmx0 !GRing.addr0
        !mul_col_row !mul0mx row_mx0 !GRing.add0r !mul0mx  !add_block_mx mul_col_mx 
  mul1mx  !GRing.addr0 !GRing.add0r mul_col_mx mul0mx row_mx0 mul_block_col 
  !mul_row_col !mulmx0 add_col_mx !GRing.addr0 add_col_mx.
move: (dba_invertible (i+1)) =>[_ ->].
move: (dba_invertiblei (i+1))  =>[-> _].
have ->: (- dab_1 (i + 1+1) *m dac (i + 1+1) + dab_1 (i + 1+1) *m dac (i +1 + 1)) = 0
    by rewrite GRing.addrC mulNmx m_m_0.
have ->: (- dcb (i+1) *m dab_1 (i+1) + dcb (i+1) *m dab_1 (i+1)) = 0
  by rewrite GRing.addrC mulNmx m_m_0.
have-> :((- dab_1 (i + 1+1) *m dac (i + 1+1) *m (- dcb (i+1) *m dab_1 (i+1)) +
         dab_1 (i + 1 + 1) *m daa (i + 1+1) + dbb (i+1) *m dab_1 (i+1)) = 0);
 last by rewrite col_mx0 -!scalar_mx_block.
rewrite !mulNmx mulmxN -mulmxA [dac (i+1+1) *m _]mulmxA d'acdca 
          mulmx_addl !mulNmx mulmx_addr !mulmxN -!mulmxA.
move : (dba_invertiblei (i+1)) => [-> _]; rewrite mulmx1.
rewrite !mulmxA.  
move : (dba_invertible (i+1)) => [ _ ->]; rewrite mul1mx.  
by rewrite GRing.addrC GRing.addrA subr_opp GRing.addrA m_m_0 GRing.sub0r  [- _ + _]GRing.addrC m_m_0.
Qed.


(*Finally, we obtain a reduction from a chain complex but it is not exactly a 
  reduction of this chain complex but the chain complex moved one dimension*)

Definition rhoHe := (Build_FGReduction (F:=fe)(G:=ge)(H:=he)fgi gf_Dh_hDie fhie hgie hhie). 



End CC_elevado.

End definitions.


End hexagonal_lemma.

